TRACE = print
local sprintf = string.format

-- Given a bunch of arguments, translates them into forms gtk_server will
-- understand and packs them into a table, which it returns
local function check_args(...)
	argv = {}
	argc = select('#', ...)
	args = { ... }
	
	for i=1,argc do
		local v = args[i]
		
		-- nil is treated as NULL
		if type(v) == 'nil' then
			argv[i] = "NULL"
			
		-- C booleans are 0 or 1
		elseif type(v) == 'boolean' then
			argv[i] = (v and "1") or "0"
		
		-- numbers - 1234 for ints, 1.234 for floats
		elseif type(v) == 'number' then
			argv[i] = sprintf((v%1 == 0 and "%d") or "%f", v)
		
		-- strings - escape ' as \', CR and LF as \r and \n, then 'singlequote'
		elseif type(v) == 'string' then
			argv[i] = sprintf([['%s']], args[i]:gsub("'", "\\'"):gsub("\n", "\\n"):gsub("\r", "\\r"))

		-- any other type is programmer error; die
		else
			error(sprintf("gtk_server argument %d: expected boolean, number, string, or nil; got %s", i, type(v)))
		end
	end
	return argv
end

-- when loaded, return a configuration function
-- this function is called with the name of the interface we want as
-- a first argument, and configuration parameters to said interface
-- (if any) as subsequent arguments

-- it subsequently returns a table (gtk_server) with the following properties:
-- gtk_server.raw(message) sends message exactly as is
-- gtk_server.cooked(...) packs its arguments into a message and sends them
-- gtk_server.gtk.foo(...) is equivalent to gtk_server("gtk_foo", ...)
-- gtk_server.gdk and gtk_server.g behave as gtk_server.gtk

-- it may also contain other, interface-specific fields useful for low level
-- access to the gtk-server
return function(interface, ...)
	local gtk_server = {}

	-- load the interface module and let it initialize
	-- it fills in some fields in the table, and also returns a sender
	-- function that we can use for blocking communication with it
	local sender = require("gtk-server."..interface)(gtk_server, ...)
	gtk_server.raw = sender

	-- fill in the subfunctions
	for i,prefix in ipairs { "gtk", "gdk", "g", "glade" } do
		gtk_server[prefix] = setmetatable({}, {
			__index = function(self, key)
				return function(...)
					return sender(prefix.."_"..key.." "..table.concat(check_args(...), " "))
				end
			end;
		})
	end

	function gtk_server.cooked(...)
		return sender(table.concat(check_args(...), " "))
	end
	gtk_server.cook_message = check_args

	package.loaded["gtk-server"] = gtk_server
	return gtk_server
end
